/*
  This file is part of SNPknock.

    Copyright (C) 2017 Matteo Sesia

    SNPknock is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SNPknock is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SNPknock.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef KNOCKOFF_MF_H
#define KNOCKOFF_MF_H

/*
MF knockoffs for a Discrete Markov Chain model
*/

#include <vector>
#include <random>

typedef std::vector< std::vector<double> > matrix;

namespace knockoffs{
  class KnockoffMF {
      int weighted_choice(double U, const std::vector<double> & weights);
  };

}

#endif
